
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "polygraph.h"

#include <iostream.h>
#include <fstream.h>

#include "Rnd.h"
#include "base/RndPermut.h"
#include "base/BStream.h"
#include "runtime/LogComment.h"
#include "runtime/ErrorMgr.h"

#include "MimeSym.h"
#include "ContentSym.h"
#include "EmbedContMdl.h"

#include "RndBodyIter.h"
#include "CdbBodyIter.h"
#include "ContainerBodyIter.h"
#include "InjectIter.h"

#include "ContentDbase.h"
#include "TextDbase.h"
#include "ObjLifeCycle.h"

#include "ContentCfg.h"

#include "runtime/polyErrors.h"


static const String DefaultContentCfgKind = "some-content";


ContentCfg::ContentCfg(int anId):
	theSize(0), theObjLifeCycle(0), theChbRatio(-1), 
	theEmbedCont(0), theCdb(0),
	theTdb(0), theInjGap(0), theInfProb(-1),
	theExtSel(0),
	theNewPerOid(0), theId(anId) {
	theObjLifeCycle = new ObjLifeCycle;
}

ContentCfg::~ContentCfg() {
	delete theObjLifeCycle;
}

void ContentCfg::configure(const ContentSym *cfg) {
	Assert(cfg);

	theKind = cfg->kind();
	if (!theKind)
		theKind = DefaultContentCfgKind;

	if (MimeSym *mime = cfg->mime()) {
		theMimeType = mime->mimeType();
		mime->extensions(theExtensions, theExtSel);
	}
	theSize = cfg->size();	
	theObjLifeCycle->configure(cfg->objLifeCycle());
	cfg->cachable(theChbRatio);

	if (cfg->hasEmbed()) {
		theEmbedCont = new EmbedContMdl;
		theEmbedCont->configure(cfg);
	}

	if (const String &cdbName = cfg->cdb()) {
		ifstream f(cdbName);
		IBStream is;
		is.configure(&f, cdbName);
		theCdb = new ContentDbase;
		theCdb->load(is);
		if (!is) {
			Comment << "error: cannot load content database from `"
				<< cdbName << "; " << error(errno) << endc;
			FatalError(errOther);
		}
	}

	if (theEmbedCont && theCdb) {
		Comment << "error: content cfg `" << theKind << "': "
			<< "cannot support containers together with content_db yet; "
			<< "do not use may_contain with content_db"
			<< endc;
		FatalError(errOther);
	}

	if (const String &tdbName = cfg->injectDb()) {
		ifstream f(tdbName);
		theTdb = new TextDbase;
		theTdb->load(f);
		if (f.bad()) {
			Comment << "error: cannot load text database from `"
				<< tdbName << "; " << error(errno) << endc;
			FatalError(errOther);
		}
		if (!theTdb->count()) {
			Comment << "error: text database `"
				<< tdbName << " appears to be empty" << endc;
			FatalError(errOther);
		}
	}

	theInjGap = cfg->injectGap();

	// XXX: we should put all inject* fields into one PGL object
	if (cfg->infectProb(theInfProb) != (theTdb != 0) || 
		(theTdb != 0) != (theInjGap != 0)) {
		Comment << "error: content cfg `" << theKind << "': "
			<< "`inject_prob' requires `inject_src' requires `inject_gap'"
			<< " and vice versa" 
			<< endc;
		FatalError(errOther);
	}
}

const String &ContentCfg::ext(int seed) const {
	static String noExt = 0;
	if (const int count = theExtensions.count()) {
		theExtSel->rndGen()->seed(seed);
		const int idx = (int)theExtSel->trial();
		Assert(0 <= idx && idx < count);
		return *theExtensions[idx];
	}
	return noExt;
}

double ContentCfg::compContPerCall(const ContentCfg *cc) const {
	if (cc->id() == id())
		return 1.0;

	if (theEmbedCont)
		return theEmbedCont->compContPerCall(cc);

	return 0.0;
}

void ContentCfg::noteNewContProb(ContentCfg *cc, double newProb) {
	if (cc->id() == id())
		return;

	if (theEmbedCont)
		theEmbedCont->noteNewContProb(cc, newProb);
}

void ContentCfg::newPerOid(double aNewPerOid) {
	Assert(aNewPerOid > 0);
	theNewPerOid = aNewPerOid;
}

// XXX: iterators should be farmed, but it is hard because they
// come in different types (perhaps somebody else should farm them?)
BodyIter *ContentCfg::getBodyIter(const ObjId &oid) {
	if (theCdb) {
		CdbBodyIter *i = new CdbBodyIter;
		i->oid(oid);
		i->cdb(theCdb);

		if (theTdb) {
			RndGen rng(RndPermut(oid.hash(), rndInjProb));
			// should we inject this object?
			if (rng.event(theInfProb)) {
				InjectIter *inj = new InjectIter; // XXX: Farm these!
				inj->creator(this);
				inj->textDbase(theTdb);
				inj->gap(theInjGap);
				i->injector(inj);
			}
		}

		return i;
	} else
	if (theEmbedCont) {
		ContainerBodyIter *i = new ContainerBodyIter;
		i->oid(oid);
		i->embedContModel(theEmbedCont);
		return i;
	} else {
		RndBodyIter *i = new RndBodyIter;
		i->oid(oid);
		return i;
	}	
}

void ContentCfg::putBodyIter(BodyIter *i) {
	i->stop();
	delete i;
}

void ContentCfg::putInjector(InjectIter *i) {
	delete i; // XXX: Farm these?
}
