
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "pgl.h"

#include <strstream.h>

#include "String.h"
#include "PglArraySym.h"
#include "PglNetAddrSym.h"

#include "PglNetAddrRange.h"
#include "gadgets.h"


/* PglNetAddrRange::Block */

PglNetAddrRange::Block::Block() {
	min = max = -1;
	pos = 0;
}

bool PglNetAddrRange::Block::valid() const {
	if (min < 0 || max < 0)
		return false;
	if (min > 0xFFFF || max > 0xFFFF)
		return false;
	if (min && min % 255 == 0)
		return false;
	if (max && max % 255 == 0)
		return false;
	if (min > max)
		return false;
	return true;
}


/* PglNetAddrRange */

PglNetAddrRange::PglNetAddrRange(): theSubnet(-1), thePort(-1),
	theRangeCnt(0), theSpillCnt(0) {
}

PglNetAddrRange::~PglNetAddrRange() {
	while(theStrs.count()) delete theStrs.pop();
}

void PglNetAddrRange::netmask(struct in_addr &netm) const {
	int subnet = theSubnet; // explicitly specified subnet, if any

	if (subnet < 0)
		subnet = (theSpillCnt > 0 || theRangeCnt > 1) ? 16 : 24;

	netm.s_addr = INADDR_NONE >> (32-subnet);
}

ArraySym *PglNetAddrRange::toSyms() {
	ArraySym *arr = new ArraySym("addr");

	startIter();
	do {
		String str;
		currentIterAddr(str);
		NetAddrSym nas;
		nas.val(NetAddr(str, thePort));
		arr->add(nas);
	} while (nextIter());

	return arr;
}

void PglNetAddrRange::toAddrs(Array<NetAddr*> &addrs) {
	startIter();
	do {
		String str;
		currentIterAddr(str);
		addrs.append(new NetAddr(str, thePort));
	} while (nextIter());
}

void PglNetAddrRange::startIter() {
	for (int i = 0; i < theBlocks.count(); ++i)
		theBlocks[i].pos = theBlocks[i].min;
}

bool PglNetAddrRange::nextIter() {
	return nextIter(theBlocks.count()-1);
}

bool PglNetAddrRange::nextIter(int level) {
	if (level < 0)
		return false;

	Block &b = theBlocks[level];
	// delegate to next level if overflow
	if (b.pos >= b.max) {
		b.pos = b.min;
		return nextIter(level-1);
	}

	b.pos++;

	if (b.pos % 255 == 0) // skip 255s
		b.pos++;

	return true;
}

void PglNetAddrRange::currentIterAddr(String &str) {
	char buf[4*4];
	ostrstream os(buf, sizeof(buf));
	for (int i = 0; i < theBlocks.count(); ++i) {
		if (i)
			os << '.';
		const Block &b = theBlocks[i];
		if (b.max > 255) // spill
			os << (b.pos / 255) << '.';
		os << (b.pos % 255);		
	}
	os << ends;
	str = buf;
}

bool PglNetAddrRange::parseBlock(const String &val) {
	Block b;
	if (const char *p = val.chr('-')) {
		if (!(isInt(val, b.min, &p) && *p == '-' && isInt(p+1, b.max, &p) && !*p))
			return false;
		if (b.max > 255)
			theSpillCnt++;
		theRangeCnt++;
	} else {
		if (!isInt(val, b.min, &p) || *p)
			return false;
		b.max = b.min;
	}

	if (b.valid()) {
		theBlocks.append(b);
		return true;
	}
	return false;
}

bool PglNetAddrRange::parse(const String &val) {
	String v = val;

	// take the port out, if any
	if (const char *p = val.chr(':')) {
		// check that subnet does not follow
		if (strchr(p, '/'))
			return false;
		v = val(0, p-val);
		if (!isInt(p+1, thePort))
			return false;
	}

	// take subnet out, if any
	if (const char *p = val.chr('/')) {
		v = val(0, p-val);
		if (!isInt(p+1, theSubnet))
			return false;
		if (!(16 <= theSubnet && theSubnet <= 32))
			return false;
	}

	// split into tokens
	while (v) {
		String token;
		if (const char *h = strchr(v, '.')) {
			token = v(0, h-v);
			v = v(h-v+1, v.len());
		} else {
			token = v;
			v = 0;
		}
		theStrs.append(new String(token));
	}

	if (theStrs.count() > 4 || theStrs.count() == 1)
		return false;

	// check that each block has valid syntax (int or int-int)
	// and form integer limits for each block
	for (int i = 0; i < theStrs.count(); ++i) {
		if (!parseBlock(*theStrs[i]))
			return false;
	}

	// check that spilling blocks have place to spill
	// any spill occupies exactly two blocks
	if (2*theSpillCnt + (theStrs.count()-theSpillCnt) > 4)
		return false;

	return true;
}

