
/*
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. All rights reserved. See the COPYRIGHT
 *  file for details.
 *
 */

#include "pgl.h"

#include <strstream.h>

#include "rndDistrs.h"
#include "popDistrs.h"

#include "PglCtx.h"
#include "PglSemx.h"
#include "PglParser.h"

#include "PglKnownNamesSym.h"
#include "PglDistrSym.h"
#include "PglPopDistrSym.h"
#include "PglRateSym.h"
#include "PglNetAddrRange.h"

#include "CacheSym.h"
#include "ContentSym.h"
#include "DumperSym.h"
#include "GoalSym.h"
#include "BenchSym.h"
#include "MimeSym.h"
#include "ObjLifeCycleSym.h"
#include "RobotSym.h"
#include "SocketSym.h"
#include "PhaseSym.h"
#include "ProxySym.h"
#include "PopModelSym.h"
#include "StatsSampleSym.h"
#include "ServerSym.h"
#include "NetPipeSym.h"
#include "UniqIdSym.h"

#include "pglStrIs.h"
#include "loadTblDistr.h"
#include "gadgets.h"

Array<AgentSym*> PglSemx::TheAgentsToUse;
Array<NetPipeSym*> PglSemx::TheNetPipesToUse;
Array<PhaseSym*> PglSemx::TheSchedule;
Array<StatsSampleSym*> PglSemx::TheSmplSchedule;
Time PglSemx::TheWorkSetLen;
int PglSemx::TheWorkSetCap = -1;

String PglSemx::TheCtxName = 0;
PglCtx *PglSemx::TheCtx = 0;

static void CheckArgs(const TokenSym &name, int expCount, const ListSym &args);
static void CheckArg(const TokenSym &name, int idx, const ListSym &args, const String &type);
static void Args2Double(const String &name, const ListSym &ina, Array<double> &outa);


/* initialization */

void PglSemx::Init() {
	Assert(!TheCtx);
	TheCtx = PglCtx::RootCtx();
	Assert(TheCtx);

	// predefined symbols
	//Declare("SideConfig", "SideCfg", TokenLoc());
}

void PglSemx::Done() {
	delete TheCtx;
	TheCtx = 0;
}

template <class Item>
static
void PglSemxDestroyArray(Array<Item> &arr) {
	while (arr.count()) delete arr.pop();
}

void PglSemx::Destroy() {
	PglSemxDestroyArray(TheAgentsToUse);
	PglSemxDestroyArray(TheNetPipesToUse);
	PglSemxDestroyArray(TheSchedule);
	PglSemxDestroyArray(TheSmplSchedule);
}


/* reduction rules */

SynSym *PglSemx::mkCode(StatementSeqSym &) { return new CodeSym; }
SynSym *PglSemx::mkStatementSeq() {	return new StatementSeqSym; }

SynSym *PglSemx::mkStatementSeq(StatementSym&, StatementSeqSym&) {
	return new StatementSeqSym;
}

SynSym *PglSemx::mkStatement(SimpleStatementSym &, TokenSym &) { 
	TheCtxName = 0;
	return new StatementSym; 
}

SynSym *PglSemx::mkStatement(BlockStatementSym &) { return new StatementSym; }

SynSym *PglSemx::mkSimpleStatement() { return new SimpleStatementSym; }
SynSym *PglSemx::mkSimpleStatement(SynSym &) { return new SimpleStatementSym; }
SynSym *PglSemx::mkExprStatement(SynSym &) { return new ExprStatementSym; }
SynSym *PglSemx::mkDeclStatement(SynSym &) { return new DeclStatementSym; }

SynSym *PglSemx::mkBlockStatement(BlockOpenSym &, SynSym &, BlockCloseSym &) {
	return new BlockStatementSym;
}

SynSym *PglSemx::mkPureDecl(TypeSym &type, CtxNameSym &name) {
	if (strchr(name.spelling(), '.'))
		cerr << name.loc() << "declaration of a record member" << endl << xexit;

	SynSymTblItem *i;
	if (TheCtx->find(name.spelling(), i)) {
		if (i->ctx() != TheCtx) {
			if (i->loc()) {
				cerr << type.loc() << "warning: local `" << name.spelling() << "' shadows earlier declaration" << endl;
				cerr << i->loc() << "warning: this is the shadowed declaration" << endl;
			} else {
				cerr << type.loc() << "warning: local `" << name.spelling() << "' shadows internal declaration" << endl;
			}
			// not an error, continue
		} else {
			if (i->loc()) {
				cerr << type.loc() << "redeclaration of `" << name.spelling() << "';" << endl;
				cerr << i->loc() << "this is the previous declaration" << endl;
			} else {
				cerr << type.loc() << "redeclaration of internally declared `" << name.spelling() << "';" << endl;
				cerr << xexit;
			}
			cerr << xexit;
		}
	}

	i = Declare(type.spelling(), name.spelling(), type.loc());
	return new PureDeclSym(i);
}

SynSym *PglSemx::mkInitDecl(PureDeclSym &d, SynSym &, ExpressionSym &e) {
	Assign(d.theDecl, e);
	return new InitDeclSym;
}

SynSym *PglSemx::mkAssignment(KnownCtxNameSym &name, SynSym &, ExpressionSym &e) {
	SynSymTblItem *i = FindDescr(name);
	Assign(i, e);
	return new AssignmentSym;
}

SynSym *PglSemx::mkAssignment(TokenSym &, KnownNamesSym &names, TokenSym &, TokenSym &, ExpressionSym &e) {
	if (names.itemType() != "KnownCtxName")
		cerr << names.loc() << "only variable names are allowed as items "
			<< "of the array on the left side of a group assignment" << endl << xexit;

	if (e.isA("Array")) {
		const ArraySym &arr = (ArraySym&)e.cast("Array");

		/* spread RHS across LHS */

		// build a random index 
		Array<int> index(arr.count());
		for (int i = 0; i < arr.count(); ++i)
			index.append(i);
		static RndGen rng;
		for (int i = 0; i < index.count(); ++i)
			index.swap(i, rng(0,index.count()));

		//if (names.count() != arr.count())
		//	cerr << names.loc() << "mismatch in the array sizes in group assignment: "
		//		<< names.count() << " != " << arr.count() << endl << xexit;

		// get LHS probabilities		
		Array<double> probs(names.count());
		names.copyProbs(probs);

		// assign each LHS item an appropriate size range from RHS
		// ranges are randomized using index built above
		for (int i = 0, idx = 0; i < names.count(); ++i) {
			Assert(names[i]);
			ArraySym iarr(arr.itemType());


			const KnownCtxNameSym &name = (const KnownCtxNameSym &)
				names[i]->cast("KnownCtxName");
			SynSymTblItem *ni = FindDescr(name);

			const int addCnt = (int)ceil(probs[i]*arr.count());
			for (; iarr.count() < addCnt && idx < index.count(); ++idx) {
				const int pos = index[idx];
				Assert(arr[pos]);
				const ExpressionSym &e = (ExpressionSym&)arr[pos]->cast("Expression");
				iarr.add(e);
			}
			Assign(ni, iarr);
		}
	} else {
		for (int i = 0; i < names.count(); ++i) {
			Assert(names[i]);
			const KnownCtxNameSym &name = (const KnownCtxNameSym &)
				names[i]->cast("KnownCtxName");
			SynSymTblItem *ni = FindDescr(name);
			Assign(ni, e);
		}
	}
	return new AssignmentSym;
}

// opens new context
SynSym *PglSemx::mkBlockOpen(TokenSym &) {
	TheCtx = new PglCtx(TheCtxName, TheCtx);
	return new BlockOpenSym;
}

// closes current context, all non-allocated symbols are destroyed
SynSym *PglSemx::mkBlockClose(TokenSym &) {
	PglCtx *cur = TheCtx;
	TheCtx = TheCtx->parent();
	Assert(TheCtx);
	//cur->report(cerr << here << "closing:" << endl);
	delete cur;
	return new BlockCloseSym;
}

SynSym *PglSemx::mkKnownCtxName(CtxNameSym &s) {
	FindDescr(s);
	return new KnownCtxNameSym(s.spelling());
}

SynSym *PglSemx::mkCtxName(ObjNameSym &s) {
	TheCtxName = s.spelling();
	return new CtxNameSym(TheCtxName);
}

SynSym *PglSemx::mkObjName(TokenSym &s) {
	return new ObjNameSym(s.spelling());
}

SynSym *PglSemx::mkObjName(TokenSym &px, TokenSym &dot, ObjNameSym &sx) {
	String name = px.spelling();
	name += dot.spelling();
	name += sx.spelling();
	return new ObjNameSym(name);
}

// general case (all processing is already done at lower levels)
SynSym *PglSemx::mkExpression(ExpressionSym &s) {
	return Any2Any(s, "Expression");
}

// 'special constant' (singly quoted string)
SynSym *PglSemx::mkExpression(TokenSym &s) {
	Assert(s.id() == SQW_STR_TOKEN);
	if (strchr(s.spelling(), '/'))
		return mkTime(s);
	else
	if (strchr(s.spelling(), '-'))
		return mkAddrRange(s);
	else
		return mkAddr(s);
}

// variable name
SynSym *PglSemx::mkExpression(ObjNameSym &name) {
	return Any2Any(FindSym(name), "Expression");
}

// { code }
SynSym *PglSemx::mkExpression(BlockStatementSym &) {
	return new NopExprSym;
}

// type { code }
SynSym *PglSemx::mkExpression(TypeSym &, BlockStatementSym &s) {
	return Any2Any(s, "Expression");
}

// ( expression ) or [ array ]
SynSym *PglSemx::mkExpression(TokenSym &, SynSym &s, TokenSym &) {
	return Any2Any(s, "Expression");
}

// op expr
SynSym *PglSemx::mkExpression(UnOperSym &op, ExpressionSym &exp) {
	return exp.unOper(op.spelling());
}

// expr op expr
SynSym *PglSemx::mkExpression(ExpressionSym &exp1, TokenSym &op, ExpressionSym &exp2) {

	// check for various exceptions and auto-conversions

	if (op.id() == CLONE_TOKEN)
		return Clone(exp1, exp2);

	// bool op
	if (op.id() == BOOL_AND_TOKEN || op.id() == BOOL_OR_TOKEN || op.id() == BOOL_XOR_TOKEN) {
		if (BoolSym *e1 = (BoolSym*)exp1.clone("bool")) {
			if (BoolSym *e2 = (BoolSym*)exp2.clone("bool")) {
				e1->loc(exp1.loc());
				e2->loc(exp2.loc());
				SynSym *res = e1->bnOper(op.spelling(), *e2);
				delete e1;
				delete e2;
				return res;
			}
			NoCast(exp2, "bool");
		}
		NoCast(exp1, "bool");
	}

	// i op f = f(i) op f
	if (exp1.canBe("int") && !exp2.canBe("int") && exp2.canBe("float")) {
		NumSym *sf = (NumSym*)exp1.clone("float");
		Assert(sf);
		sf->loc(exp1.loc());
		SynSym *res = sf->bnOper(op.spelling(), exp2);
		delete sf;
		return res;
	}

	/* rates and such */
	if (op.id() == DIV_TOKEN && exp2.canBe("time")) {
		// i,f / t = rate
		if (NumSym *f = (NumSym*)exp1.clone("float")) {
			TimeSym *time = (TimeSym*)exp2.clone("time");
			SynSym *rate = new RateSym(f->val(), time->val());
			delete time;
			delete f;
			return rate;
		}

		/*
		// size / t = bandwidth
		if (exp1.isA("size")) {
			// return new BwidthSym(e1.val(), e2.val());
		}*/
	}

	if (op.id() == DIV_TOKEN && exp2.canBe("rate")) {
		// i,f / rate = t
		if (NumSym *f = (NumSym*)exp1.clone("float")) {
			RateSym *rate = (RateSym*)exp2.clone("rate");
			SynSym *time = new TimeSym(Time::Secd(f->val() / rate->val()));
			delete rate;
			delete f;
			return time;
		}

		/*
		// size / t = bandwidth
		if (exp1.isA("size")) {
			// return new BwidthSym(e1.val(), e2.val());
		}*/
	}

	if (op.id() == MUL_TOKEN) {
		// i,f * t,s,r =  t,s,r * i,f
		if (exp1.canBe("float") &&
			(exp2.canBe("time") || exp2.canBe("size") || exp2.canBe("rate")))
			return exp2.bnOper(op.spelling(), exp1);
	}

	// default: exp1 will handle the op
	return exp1.bnOper(op.spelling(), exp2);
}

// f(expr1, expr2, ...)
SynSym *PglSemx::mkExpression(TokenSym &fname, TokenSym &, ListSym &args, TokenSym &) {

	if (!strcmp(fname.spelling(), "int")) {
		CheckArgs(fname, 1, args);
		return Any2Int(*args[0]);
	} else
	if (!strcmp(fname.spelling(), "float")) {
		CheckArgs(fname, 1, args);
		return Any2Float(*args[0]);
	} else
	if (!strcmp(fname.spelling(), "undef")) {
		CheckArgs(fname, 0, args);
		return new UndefOpExprSym();
	} else
	if (!strcmp(fname.spelling(), "count")) {
		CheckArgs(fname, 1, args);
		CheckArg(fname, 0, args, "List");
		return new IntSym(((ListSym&)args[0]->cast("List")).count());
	} else
	if (!strcmp(fname.spelling(), "max")) {
		return CalcExtreme(fname, args, +1);
	} else
	if (!strcmp(fname.spelling(), "min")) {
		return CalcExtreme(fname, args, -1);
	} else
	if (!strcmp(fname.spelling(), "uniqId")) {
		CheckArgs(fname, 0, args);
		return new UniqIdSym(UniqId::Create());
	} else
	if (!strcmp(fname.spelling(), "clientHostCount")) {
		CheckArgs(fname, 1, args);
		CheckArg(fname, 0, args, "Bench");
		if (IntSym *is = ((BenchSym&)args[0]->cast("Bench")).clientHostCount())
			return is;
		cerr << fname.loc() << "clientHostCount(): cannot compute "
			<< "number of client hosts using the given bench configuration; "
			<< "are peak_req_rate and max_client_load set?" << endl << xexit;
	} else
	if (!strcmp(fname.spelling(), "robotAddrs")) {
		CheckArgs(fname, 1, args);
		CheckArg(fname, 0, args, "Bench");
		if (ArraySym *as = ((BenchSym&)args[0]->cast("Bench")).robots())
			return as;
		cerr << fname.loc() << "invalid bench configuration "
			<< "for the robotIps() function" << endl << xexit;
	} else
	if (!strcmp(fname.spelling(), "serverAddrs")) {
		CheckArgs(fname, 1, args);
		CheckArg(fname, 0, args, "Bench");
		if (ArraySym *as = ((BenchSym&)args[0]->cast("Bench")).servers())
			return as;
		cerr << fname.loc() << "invalid bench configuration "
			<< "for the serverIps() function" << endl << xexit;
	} else
	if (!strcmp(fname.spelling(), "place")) {
		cerr << fname.loc() << "place() function is no longer supported;"
			<< " just set the `hosts' field to the list of addresses intead"
			<< endl << xexit;
	} else
	if (SynSym *ds = isDistr(fname, args))
		return ds;
	else
	if (SynSym *pms = isPopDistr(fname, args))
		return pms;

	cerr << fname.loc() << "unknown function, distribution, or popularity model `" << fname.spelling() << "'" << endl << xexit;
	return 0;
}

// p(expr1, expr2, ...)
SynSym *PglSemx::mkCall(TokenSym &pname, TokenSym &, ListSym &args, TokenSym &) {
	if (!strcmp(pname.spelling(), "use")) {
		Use(args);
	} else
	if (!strcmp(pname.spelling(), "schedule")) {
		Schedule(args);
	} else
	if (!strcmp(pname.spelling(), "working_set_length")) {
		CheckArgs(pname, 1, args);
		CheckArg(pname, 0, args, "time");
		TheWorkSetLen = ((TimeSym&)args[0]->cast("time")).val();
	} else
	if (!strcmp(pname.spelling(), "working_set_cap")) {
		CheckArgs(pname, 1, args);
		CheckArg(pname, 0, args, "float");
		TheWorkSetCap = (((IntSym&)args[0]->cast("int")).val());
	} else {
		cerr << pname.loc() << "unknown procedure call `" << pname.spelling() << "'" << endl << xexit;
		return 0;
	}
	return new CallSym();
}

SynSym *PglSemx::mkUnOper(TokenSym &s) {
	return new UnOperSym(s.spelling());
}

SynSym *PglSemx::mkType(TokenSym &type) { 
	/*static int anonCnt = 0;
	strstream s;
	s << type.spelling() << "_anonym_" << ++anonCnt << ends;
	TheCtxName = s.str();
	s.freeze(false);
	Declare(type.spelling(), TheCtxName, type.loc());
	*/
	return new TypeSym(type.spelling()); 
}

// name[]
SynSym *PglSemx::mkType(TokenSym &s, TokenSym &, TokenSym &) { 
	return new TypeSym(s.spelling() + String("[]")); 
}

SynSym *PglSemx::mkBool(TokenSym &s) {
	Assert(s.id() == BOOL_TRUE_TOKEN || s.id() == BOOL_FALSE_TOKEN);
	return new BoolSym(s.id() == BOOL_TRUE_TOKEN);
}

SynSym *PglSemx::mkInt(TokenSym &s) {
	int v;
	Assert(isInt(s.spelling(), v));
	return new IntSym(v);
}

SynSym *PglSemx::mkNum(TokenSym &s) {
	double v;
	const char *p = 0;
	Assert(isNum(s.spelling(), v, &p));
	if (p && *p)
		v /= 100;
	return new NumSym(v);
}

SynSym *PglSemx::mkTime(TokenSym &s) {
	Time v;
	if (s.id() == TIME_TOKEN)
		Assert(pglIsRelTime(s.spelling(), v));
	else
	if (!pglIsAbsTime(s.spelling(), v))
		cerr << s.loc() << "malformed time constant: '" << s.spelling() << "'" << endl << xexit;
	return new TimeSym(v);
}

SynSym *PglSemx::mkSize(TokenSym &s) {
	Assert(s.id() == SIZE_TOKEN);
	BigSize v;
	Assert(pglIsSize(s.spelling(), v));
	return new SizeSym(v);
}

SynSym *PglSemx::mkString(TokenSym &s) {
	return new StringSym(s.spelling());
}

SynSym *PglSemx::mkAddr(TokenSym &s) {
	String host;
	int port;
	if (pglIsNetAddr(s.spelling(), host, port)) {
		NetAddrSym *addr = new NetAddrSym();
		addr->val(NetAddr(host, port));
		return addr;
	}

	cerr << s.loc() << "malformed address constant: '" << s.spelling() << "'" << endl << xexit;
	return 0;
}

SynSym *PglSemx::mkAddrRange(TokenSym &s) {
	PglNetAddrRange ar;
	if (ar.parse(s.spelling()))
		return ar.toSyms();

	cerr << s.loc() << "malformed address range constant: '" << s.spelling() << "'" << endl << xexit;
	return 0;
}

SynSym *PglSemx::mkQualif(TokenSym &s) {
	if (s.spelling() == "lmt")
		return new QualifSym(QualifSym::qfLmt);
	if (s.spelling() == "now")
		return new QualifSym(QualifSym::qfNow);
	if (s.spelling() == "nmt")
		return new QualifSym(QualifSym::qfNmt);
	Assert(false);
	return 0;
}

// [ name ]
SynSym *PglSemx::mkKnownNames(KnownCtxNameSym &name) {
	KnownNamesSym *names = new KnownNamesSym("KnownCtxName");
	static KnownNamesSym *NN;
	NN = names;
	names->add(name);
	static SynSym *NS;
	NS = names;
	static KnownNamesSym *NN2;
	NN2 = (KnownNamesSym*) NS;
	return names;
}

// [ name : p ]
SynSym *PglSemx::mkKnownNames(KnownCtxNameSym &name, TokenSym &, ExpressionSym &p) {
	NumSym *prob = Any2Float(p);
	KnownNamesSym *names = new KnownNamesSym("KnownCtxName");
	names->addProb(name, prob->val());
	delete prob;
	return names;
}

// [ names, name ]
SynSym *PglSemx::mkKnownNames(KnownNamesSym &src, TokenSym &, KnownCtxNameSym &name) {
	KnownNamesSym *names = (KnownNamesSym*) src.dup();
	Assert(names);
	names->add(name);
	return names;
}

// [ names, name : p ]
SynSym *PglSemx::mkKnownNames(KnownNamesSym &arr, TokenSym &, KnownCtxNameSym &name, TokenSym &, ExpressionSym &p) {
	NumSym *prob = Any2Float(p);
	KnownNamesSym *names = (KnownNamesSym*) arr.dup();
	Assert(names);
	names->addProb(name, prob->val());
	delete prob;
	return names;
}

// []
SynSym *PglSemx::mkArray() {
	return new ArraySym(0); // anonymous-type array
}

// [ exp ]
SynSym *PglSemx::mkArray(ExpressionSym &exp) {
	ArraySym *a = new ArraySym(0);
	a->add(exp);
	return a;
}

// [ exp : p ]
SynSym *PglSemx::mkArray(ExpressionSym &exp, TokenSym &, ExpressionSym &p) {
	NumSym *prob = Any2Float(p);
	ArraySym *a = new ArraySym(0);
	a->addProb(exp, prob->val());
	delete prob;
	return a;
}

// [ arr, exp ]
SynSym *PglSemx::mkArray(ArraySym &arr, TokenSym &, ExpressionSym &exp) {
	ArraySym *a = (ArraySym*) arr.dup();
	Assert(a);
	a->add(exp);
	return a;
}

// [ arr, exp : p ]
SynSym *PglSemx::mkArray(ArraySym &arr, TokenSym &, ExpressionSym &exp, TokenSym &, ExpressionSym &p) {
	NumSym *prob = Any2Float(p);
	ArraySym *a = (ArraySym*) arr.dup();
	Assert(a);
	a->addProb(exp, prob->val());
	delete prob;
	return a;
}

SynSym *PglSemx::mkList() {
	return new ListSym(); // empty list
}

SynSym *PglSemx::mkList(ExpressionSym &exp) {
	ListSym *a = new ListSym();
	a->add(exp);
	return a;
}

SynSym *PglSemx::mkList(ListSym &list, TokenSym &, ExpressionSym &exp) {
	ListSym *l = (ListSym*) list.dup();
	Assert(l);
	l->add(exp);
	return l;
}



/* tools */

// finds corresponding description or exits
SynSymTblItem *PglSemx::FindDescr(const TokenSym &name) {
	SynSymTblItem *i;
	if (TheCtx->find(name.spelling(), i)) {
		Assert(i);
		return i;
	}

	cerr << name.loc();
	if (strchr(name.spelling(), '.'))
		cerr << "`" << name.spelling() << "' or its component is not declared" << endl << xexit;
	else
		cerr << "`" << name.spelling() << "' is not declared" << endl << xexit;
	cerr << xexit;
	return 0;
}

// finds corresponding (defined) symbol or exits
SynSym &PglSemx::FindSym(const TokenSym &name) {
	SynSymTblItem *i = FindDescr(name);
	if (!i->sym())
		cerr << name.loc() << "`" << name.spelling() << "' is not defined" << endl << xexit;
	return *i->sym();
}

SynSymTblItem *PglSemx::Declare(const String &type, const String &name, const TokenLoc &loc) {
	SynSymTblItem *i = new SynSymTblItem(type, name);
	i->loc(loc);
	TheCtx->add(name, i);
	SetDefault(i);
	//if (i->sym()) cerr << here << i->sym()->type() << " : " << *i->sym() << endl;
	return i;
}

void PglSemx::Assign(SynSymTblItem *i, const ExpressionSym &e) {
	// if (i->sym()) cerr << here << i->name() << " was: (" << i->sym()->type() << ") " << *i->sym() << endl;
	// cerr << here << i->name() << " = (" << e.type() << ") " << e << endl;

	if (e.isA("NopExpr")) // no change requested
		return;

	Assert(i->sym() != &(SynSym&)e);
	delete i->sym();            // destroy previous value
	i->sym(0);

	if (e.isA("UndefOpExpr"))   // undefine
		return;

	i->sym(e.clone(i->type())); // set new one

	if (!i->sym()) {
		cerr << e.loc() << "assignment of incompatible type; "
			<< "no default conversion from `" << e.type()
			<< "' to `" << i->type() << "'" << endl << xexit;
	}
}

SynSym *PglSemx::Clone(SynSym &s, ExpressionSym &exp) {
	if (IntSym *expi = (IntSym *) exp.clone("int")) {
		if (expi->val() >= 0) {
			ArraySym *a = new ArraySym(0);
			for (int i = 0; i < expi->val(); ++i)
				a->add(s);
			return a;
		}
		cerr << exp.loc() << "negative cloning factor (" << expi->val() << ")" << endl << xexit;
	}
	cerr << exp.loc() << "cloning factor must be an integer expression" << endl;
	NoCast(exp, "int");
	return 0;
}

SynSym *PglSemx::Any2Any(const SynSym &s, const String &type) {
	if (SynSym *clone = s.clone(type))
		return clone;
	NoCast(s, type);
	return 0;
}

IntSym *PglSemx::Any2Int(const SynSym &s) {
	if (s.isA("float")) {
		const NumSym &f = (const NumSym&)s.cast("float");
		return IntSym::Fit("conversion to int", f.val(), s.loc());
	}

	if (IntSym *i = (IntSym*)s.clone("int")) // default casts
		return i;

	NoCast(s, "int");
	return 0;
}

NumSym *PglSemx::Any2Float(const SynSym &s) {
	if (NumSym *f = (NumSym*)s.clone("float")) // default casts
		return f;
	NoCast(s, "float");
	return 0;
}

void PglSemx::NoCast(const SynSym &s, const char *totype) {
	cerr << s.loc() << "no conversion from `" << s.type()
		<< "' to `" << totype << "'" << endl << xexit;
}

ArraySym *PglSemx::Place(const AgentSym &agent, const ArraySym &addrs) {
	ArraySym *arr = new ArraySym(agent.type());
	for (int i = 0; i < addrs.count(); ++i) {
		AgentSym *a = (AgentSym*) agent.dup();
		SynSymTblItem *host = 0;
		Assert(a->rec());
		Assert(a->rec()->find("host", host));
		Assert(host);
		Assign(host, (NetAddrSym&)addrs[i]->cast("addr"));
		arr->add(*a);
		delete a;
	}
	return arr;
}

// register agents that will be actually used
void PglSemx::Use(const ListSym &objects) {
	for (int i = 0; i < objects.count(); ++i) {
		if (objects[i]->isA("List"))
			Use((ListSym &)objects[i]->cast("List")); // flatten
		else
		if (AgentSym *a = (AgentSym*)objects[i]->clone("Agent"))
			TheAgentsToUse.append(a);
		else
		if (NetPipeSym *p = (NetPipeSym*)objects[i]->clone("NetPipe"))
			TheNetPipesToUse.append(p);
		else
			cerr << objects[i]->loc() << "entry of type `" << objects[i]->type() << "' in `use()' argument list" << endl << xexit;
	}
}

// register stat phases that will be actually used
void PglSemx::Schedule(const ListSym &items) {
	for (int i = 0; i < items.count(); ++i) {
		if (items[i]->isA("List"))
			Schedule((ListSym &)items[i]->cast("List")); // flatten
		else
		if (PhaseSym *p = (PhaseSym*)items[i]->clone("Phase"))
			TheSchedule.append(p);
		else
		if (StatsSampleSym *s = (StatsSampleSym*)items[i]->clone("StatsSample"))
			TheSmplSchedule.append(s);
		else
			cerr << items[i]->loc() << "entry of type `" << items[i]->type() << "' cannot be scheduled" << endl << xexit;
	}
}

SynSym *PglSemx::CalcExtreme(const TokenSym &name, const ListSym &args, int dir) {
	if (!args.count())
		cerr << name.loc() << "max() and min() functions need at least one argument" << endl << xexit;

	// flatten if needed
	if (args.count() == 1 && args[0]->isA("List"))
		return CalcExtreme(name, (ListSym &)args[0]->cast("List"), dir);

	// convert all arguments to double (ideally, we should use args directly)
	Array<double> dargs;
	Args2Double(name.spelling(), args, dargs);

	// find position of an extreme
	int pos = 0;
	for (int i = 1; i < dargs.count(); ++i) {
		if (dir*dargs[i] > dir*dargs[pos])
			pos = i;
	}

	return args[pos]->dup();
}

// check argument type
static
void CheckArg(const TokenSym &name, int idx, const ListSym &args, const String &type) {
	Assert(idx < args.count());
	
	if (!args[idx]->canBe(type)) {
		cerr << name.loc() << "bad argument " << idx+1 << " for `" << name.spelling() 
			<< "'; type `" << type << "' expected, got `" << args[idx]->type() << "'"
			<< endl << xexit;
	}
}

// check the number of arguments
static
void CheckArgs(const TokenSym &name, int expCount, const ListSym &args) {
	if (args.count() != expCount) {
		cerr << name.loc() << "`" << name.spelling() 
			<< "' expected " << expCount << " arguments, got " << args.count()
			<< endl << xexit;
	}

	// sanity check: all args should be defined
	for (int i = 0; i < args.count(); ++i)
		Assert(args[i]);
}

// convert all arguments to double (including Size and Time)
static
void Args2Double(const String &name, const ListSym &ina, Array<double> &outa) {
	Assert(!outa.count());
	for (int i = 0; i < ina.count(); ++i) {
		const SynSym &s = *ina[i];
		if (NumSym *n = (NumSym *)s.clone("float")) {
			outa.append(n->val());
			delete n;
		} else
		if (TimeSym *t = (TimeSym*) s.clone("time")) {
			outa.append(t->val().secd());
			delete t;
		} else
		if (SizeSym *sz = (SizeSym*) s.clone("size")) {
			outa.append(sz->val().byted());
			delete sz;
		} else
			cerr << s.loc() << "invalid type `" << s.type()
				<< "' for argument #" << i+1
				<< " in `" << name << "'" << endl << xexit;
	}
}

// convert all arguments to int
static
void Args2Int(const String &name, const ListSym &ina, Array<int> &outa) {
	Assert(!outa.count());
	for (int i = 0; i < ina.count(); ++i) {
		const SynSym &s = *ina[i];
		if (IntSym *is = (IntSym*) s.clone("int")) {
			outa.append(is->val());
			delete is;
		} else
			cerr << s.loc() << "invalid type `" << s.type()
				<< "' for argument #" << i+1
				<< " in `" << name << "'; integer expected" << endl << xexit;
	}
}

DistrSym *PglSemx::isDistr(const TokenSym &sname, ListSym &args) {
	RndGen *gen = new RndGen;
	Array<double> dargs;
	const String name = sname.spelling();

	// table distribution requires exceptional handling, others are below
	if (name == "table") {
		CheckArgs(sname, 2, args);
		const String &fName =
			((const StringSym&)(Any2Any(*args[0], "String")->cast("String"))).val();
		const String &aType = 
			((const StringSym&)(Any2Any(*args[1], "String")->cast("String"))).val();
		if (RndDistr *d = LoadTblDistr(fName, aType))
			return new DistrSym(aType + "_distr", d);
		cerr << sname.loc() << "failed to load the table with `"
			<< aType << "' distribution from `" << fName << "'" << xexit;
	}
	
	// by default, guess type of values by the type of the first argument
	const String dType = args.count() ? args[0]->type() + "_distr" : String(0);

	// XXX: Args2Double can call exit() while is* should just return 0;
	Args2Double(name, args, dargs);

	if (name == "const") {
		CheckArgs(sname, 1, args);
		return new DistrSym(dType, new ConstDistr(gen, dargs[0]));
	} else
	if (name == "unif") {
		CheckArgs(sname, 2, args);
		return new DistrSym(dType, new UnifDistr(gen, dargs[0], dargs[1]));
	} else
	if (name == "exp") {
		CheckArgs(sname, 1, args);
		return new DistrSym(dType, new ExpDistr(gen, dargs[0]));
	} else
	if (name == "norm") {
		CheckArgs(sname, 2, args);
		return new DistrSym(dType, new NormDistr(gen, dargs[0], dargs[1]));
	} else
	if (name == "logn") {
		CheckArgs(sname, 2, args);
		return new DistrSym(dType, LognDistr::ViaMean(gen, dargs[0], dargs[1]));
	} else
	if (name == "zipf") {
		CheckArgs(sname, 1, args);
		Array<int> iargs;
		Args2Int(name, args, iargs);
		return new DistrSym(dType, new ZipfDistr(gen, iargs[0]));
	} else
	if (name == "seq") {
		CheckArgs(sname, 1, args);
		Array<int> iargs;
		Args2Int(name, args, iargs);
		return new DistrSym(dType, new SeqDistr(gen, iargs[0]));
	} else {
		delete gen;
		return 0;
	}
}

PopDistrSym *PglSemx::isPopDistr(const TokenSym &sname, ListSym &args) {
	const String name = sname.spelling();
	const String dType = "PopDistr";
	Array<double> dargs;
	Args2Double(name, args, dargs);

	if (name == "pmUnif") {
		CheckArgs(sname, 0, args);
		return new PopDistrSym(dType, new UnifPopDistr);
	} else
	if (name == "pmZipf") {
		CheckArgs(sname, 1, args);
		return new PopDistrSym(dType, new ZipfPopDistr(dargs[0]));
	}

	return 0;
}

// note: also checks if the type is a "known" type!
void PglSemx::SetDefault(SynSymTblItem *i) {

	if (!KnownType(i->type()))
		cerr << i->loc() << "unknown type `" << i->type() << "'" << endl << xexit;

	if (i->type() == "Dumper")
		i->sym(new DumperSym);
	else
	if (i->type() == "Socket")
		i->sym(new SocketSym);
	else
	if (i->type() == "Goal")
		i->sym(new GoalSym);
	else
	if (i->type() == "Agent")
		i->sym(new AgentSym("Agent"));
	else
	if (i->type() == "Server")
		i->sym(new ServerSym);
	else
	if (i->type() == "Robot")
		i->sym(new RobotSym);
	else
	if (i->type() == "Proxy")
		i->sym(new ProxySym);
	else
	if (i->type() == "PopModel")
		i->sym(new PopModelSym);
	else
	if (i->type() == "Bench")
		i->sym(new BenchSym);
	else
	if (i->type() == "Cache")
		i->sym(new CacheSym);
	else
	if (i->type() == "Mime")
		i->sym(new MimeSym);
	else
	if (i->type() == "ObjLifeCycle")
		i->sym(new ObjLifeCycleSym);
	else
	if (i->type() == "Phase")
		i->sym(new PhaseSym);
	else
	if (i->type() == "StatsSample")
		i->sym(new StatsSampleSym);
	else
	if (i->type() == "Content")
		i->sym(new ContentSym);
	else
	if (i->type() == "NetPipe")
		i->sym(new NetPipeSym);
	// else no default value
}

// checks if the type is a "known" type
bool PglSemx::KnownType(const String &type) {
	static Array<String*> knownTypes;

	// work around an internal gcc bug
	if (!knownTypes.count()) {
		// these are all known integral types
		static const char *kts[] = {
			"bool", "int", "float",
			"time", "size", "rate", "bwidth", "addr", "string",
			"uniq_id",
			"Agent", "Server", "Robot", "Proxy", 
			"PopDistr", "PopModel", "Cache",
			"Bench",
			"Mime", "ObjLifeCycle", "Content",
			"Phase", "StatsSample", "NetPipe"
		};
		const int ktCount = (int)(sizeof(kts)/sizeof(*kts));

		for (int i = 0; i < ktCount; ++i)
			knownTypes.append(new String(kts[i]));
	}

	for (int i = 0; i < knownTypes.count(); ++i) {
		const String &kt = *knownTypes[i];
		if (type == kt)
			return true;
		if (type == kt + "_distr") // distributions
			return true;
		if (type == kt + "[]")     // arrays
			return true;
		if (type == kt + "_distr[]") // arrays of distributions
			return true;
	}

	return false;
}
